function init(tilemapid,roomsize_,tilesize_)
	map = TileMap.new(tilemapid)
	roomsize = roomsize_
	tilesize = tilesize_
	colours = {}
	features = {}
	visualfeatures = {}
	featureindex = {}
	objectdata = {}
	units = {}
	tiledunits = {}
	codeunits = {}
	unitlists = {}
	undobuffer = {}
	animunits = {}
	unitmap = {}
	firstwords = {}
	deleted = {}
	currid = 0
	currdir = 4
	updatecode = 1
	doundo = true
	updateundo = true
	dokeyblock = ""
	ruledebug = false
end

function addunit(id)
	local unitid = #units + 1
	
	units[unitid] = {}
	units[unitid] = mmf.newObject(id)
	
	local unit = units[unitid]
	
	if (unit.values[ID] == -1) then
		unit.values[ID] = newid()
	end

	if (unit.values[XPOS] > 0) and (unit.values[YPOS] > 0) then
		addunitmap(id,unit.values[XPOS],unit.values[YPOS])
	end
	
	if (unit.values[TILING] == 1) then
		table.insert(tiledunits, unit.fixed)
	end
	
	if (unit.values[TILING] > 1) then
		table.insert(animunits, unit.fixed)
	end
	
	local name = unit.strings[UNITNAME]
	if (unitlists[name] == nil) then
		unitlists[name] = {}
	end
	
	table.insert(unitlists[name], unit.fixed)
	
	if (unit.strings[UNITTYPE] == "text") then
		table.insert(codeunits, unit.fixed)
		updatecode = 1
	end
end

function clearunits()
	units = {}
	tiledunits = {}
	codeunits = {}
	animunits = {}
	unitlists = {}
	undobuffer = {}
	unitmap = {}
	currid = 0
	updateundo = true
	newundo()
	currdir = 4
end

function smallclear()
	objectdata = {}
	deleted = {}
end

function clear()
	features = {}
	featureindex = {}
	visualfeatures = {}
	objectdata = {}
	firstwords = {}
	deleted = {}
	updatecode = 1
	updateundo = false
	dokeyblock = ""
end

function command(key)
	--local special = featureindex[key]
	local special = nil
	local keyid = -1
	if (keys[key] ~= nil) then
		keyid = keys[key]
	else
		print("no such key")
		return
	end

	if (special == nil) then
		if (keyid <= 4) then
			local drs = ndirs[keyid+1]
			local ox = drs[1]
			local oy = drs[2]
			local dir = keyid
			
			if (dir == 4) then
				dir = nil
			end
			
			MF_update()
			movecommand(ox,oy,dir)
		end
		
		if (keyid == 5) then
			MF_restart()
		end
	else
		MF_update()
		dokeyblock = special
	end
end

function conversion()
	for i,v in ipairs(features) do
		local words = {}
		for a,b in ipairs(v) do
			table.insert(words, b)
		end
		
		local operator = ""
		local docheck = ""
		
		if (#words == 3) then
			operator = words[2]
		elseif (#words == 5) then
			operator = words[4]
			docheck = words[3]
		end
		
		if (operator == "is") then
			local output = {}
			local v1 = words[1]
			
			local mat1 = 0
					
			if (v1 ~= "text") and (v1 ~= "level") then
				mat1 = material(v1)
			else
				mat1 = 1
			end
			
			local mat2 = 1
			
			if (mat1 == 1) then
				for a,b in ipairs(features) do
					local b1,b2,b3 = "","",""
					local docheck2 = {}
					
					if (#b == 3) then
						b1 = b[1]
						b2 = b[2]
						b3 = b[3][1]
						docheck2 = nil
					elseif (#b == 5) then
						b1 = b[1]
						b2 = b[4]
						b3 = b[5]
						docheck2 = b[3]
					end
					
					if (b1 == v1) and (b2 == "is") then
						if (b3 ~= "text") then
							for d,c in ipairs(materials) do
								if (c == b3) and (v1 ~= b3) and (c ~= "group") then
									table.insert(output, {b3, docheck2})
								end
							end
						else
							if (v1 ~= b3) then
								table.insert(output, {b3, docheck2})
							end
						end
					end
				end
				
				if (#output > 0) then
					local conversions = {}
					
					for k,v3 in ipairs(output) do
						if (v3[1] ~= "all") and (v3[1] ~= "text") and (v3[1] ~= "group") then
							table.insert(conversions, v3)
						elseif (v3[1] == "all") then
							createall({v1,v3[2]})
						elseif (v3[1] == "text") then
							table.insert(conversions, {"text_" .. v1, v3[2]})
						end
					end
					
					if (#conversions > 0) then
						if (v1 ~= "text") then
							convert(v1,conversions)
						else
							local convertthese = {}
							for a,b in ipairs(units) do
								if (b.strings[UNITTYPE] == "text") then
									table.insert(convertthese, b.strings[UNITNAME])
								end
							end
							
							if (#convertthese > 0) then
								for a,b in ipairs(convertthese) do
									convert(b,conversions)
								end
							end
						end
					end
				end
			end
		end
	end
end

function findfeature(rule1,rule2,rule3)
	local options = {}
	local result = {}
	local rule = ""
	
	if (rule1 ~= nil) then
		rule = rule1 .. " "
	end
	
	if (rule2 ~= nil) then
		rule = rule .. rule2 .. " "
	end
	
	if (rule3 ~= nil) then
		rule = rule .. rule3
	end
	
	if (featureindex[rule1] ~= nil) then
		for i,v in ipairs(featureindex[rule1]) do
			if (v[2] ~= "on") then
				if (v[1] == rule1) and (v[2] == rule2) then
					table.insert(options, {v[1],v[2],v[3][1]})
				end
			else
				if (v[1] == rule1) and (v[4] == rule2) then
					table.insert(options, {v[1],v[4],v[5],v[3]})
				end
			end
		end
	end
	
	if (featureindex[rule3] ~= nil) then
		for i,v in ipairs(featureindex[rule3]) do
			if (v[2] ~= "on") then
				if (v[3][1] == rule3) and (v[2] == rule2) then
					table.insert(options, {v[1],v[2],v[3][1]})
				end
			else
				if (v[5] == rule3) and (v[4] == rule2) then
					table.insert(options, {v[1],v[4],v[5],v[3]})
				end
			end
		end
	end
	
	if (rule1 == nil) and (rule3 == nil) and (rule2 ~= nil) then
		if (featureindex[rule2] ~= nil) then 
			for i,v in ipairs(featureindex[rule2]) do
				local usable = false
				
				if (v[2] ~= "on") then
					for a,b in ipairs(materials) do
						if (b == v[3][1]) then
							usable = true
						end
					end
					
					if (v[2] == rule2) and usable then
						table.insert(options, {v[1],v[2],v[3][1]})
					end
				else
					for a,b in ipairs(materials) do
						if (b == v[5]) then
							usable = true
						end
					end
					
					if (v[4] == rule2) and usable then
						table.insert(options, {v[1],v[4],v[5],v[3]})
					end
				end
			end
		end
	end
	
	for i,v in ipairs(options) do
		local words = {}
		local limit = 0
		
		for a,b in ipairs(v) do
			table.insert(words, b)
			limit = limit + 1
			
			if (limit == 4) then
				break
			end
		end
		
		if (#words >= 3) then
			local one = words[3]
			local two = words[2] .. " " .. words[3]
			local three = words[1] .. " " .. words[2] .. " " .. words[3]
			
			if (one == rule) or (two == rule) or (three == rule) or ((rule2 == words[2]) and (rule1 == nil) and (rule3 == nil)) then
				if (v[4] == nil) then
					table.insert(result, {v[1]})
				else
					table.insert(result, {v[1], v[4]})
				end
			end
		end
	end
	
	if (#result > 0) then
		return result
	else
		return nil
	end
end

function createall(matdata,x_,y_,id_)
	local all = {}
	local empty = false
	
	if (x_ == nil) and (y_ == nil) and (id_ == nil) then
		if (matdata[1] ~= "empty") and (matdata[1] ~= "level") and (matdata[1] ~= "group") then
			all = findall(matdata)
		else
			all = findempty()
			empty = true
		end
	end
	local test = {}
	
	if (x_ ~= nil) and (y_ ~= nil) and (id_ ~= nil) then
		local check = findtype(matdata,x_,y_,id_)
		
		if (#check > 0) then
			for i,v in ipairs(check) do
				if (v > 0) then
					table.insert(test, v)
				end
			end
		end
	end
	
	if (#all > 0) then
		for i,v in ipairs(all) do
			table.insert(test, v)
		end
	end
	
	if (#test > 0) then
		for i,v in ipairs(test) do
			if (empty == false) then
				local x,y,dir = pos(v)
						
				for a,b in ipairs(materials) do
					if (b ~= "empty") and (b ~= "all") and (b ~= "level") and (b ~= "group") then
						local mat = findtype({b},x,y,v)
						local tmat = findtext(x,y)
						
						if ((#mat == 0) and (tmat == 0)) or ((#mat == 0) and (matdata[1] == "text")) then
							create(b,x,y,dir)
							
							if (matdata[1] == "text") and (tmat > 0) then
								delete(tmat,0)
							end
						end
					end
				end
			else
				local x = v % roomsize
				local y = math.floor(v / roomsize)
				local dir = 4
				
				for a,b in ipairs(materials) do
					if (b ~= "empty") and (b ~= "all") and (b ~= "level") and (b ~= "group") then
						local mat = findtype({b},x,y,0)
						local tmat = findtext(x,y)
						
						if ((#mat == 0) and (tmat == 0)) or ((#mat == 0) and (matdata[1] == "text")) then
							create(b,x,y,dir)
							
							if (matdata[1] == "text") and (tmat > 0) then
								delete(tmat,0)
							end
						end
					end
				end
			end
		end
	end
end

function issafe(unitid)
	local unit = mmf.newObject(unitid)
	local x,y = unit.values[XPOS],unit.values[YPOS]
	local name = unit.strings[UNITNAME]
	
	if (unit.strings[UNITTYPE] == "text") then
		name = "text"
	end
	
	local safe = hasfeature(name,"is","safe",unitid)
	
	if (safe ~= nil) then
		return true
	end
	
	return false
end

function setunitmap()
	unitmap = {}
	
	for i,unit in ipairs(units) do
		local tileid = unit.values[XPOS] + unit.values[YPOS] * roomsize
		
		--print(tostring(unit.values[XPOS]) .. ", " .. tostring(unit.values[YPOS]) .. ", " .. unit.strings[UNITNAME])
		
		if (unitmap[tileid] == nil) then
			unitmap[tileid] = {}
		end

		table.insert(unitmap[tileid], unit.fixed)
	end
end

function setundo(this)
	if (this ~= nil) then
		if (this == 1) then
			updateundo = true
		elseif (this == 0) then
			updateundo = false
		end
	else
		print("undo is nil!")
		updateundo = true
	end
end

function victory()
	MF_win()
end